@@ -2,11 +2,12 @@
 import ctypes
 import os
 import time
-from typing import Tuple
 
 import numpy as np
 import tensorflow as tf
 
+from openke.utils import generate_type_constraint_specification
+
 
 class Config(object):
     '''
@@ -14,7 +15,7 @@ class Config(object):
     '''
 
     def __init__(self):
-        base_file = os.path.abspath(os.path.join(os.path.dirname(__file__), '../release/Base.so'))
+        base_file = "/usr/lib/libmeager.so"
         self.lib = ctypes.cdll.LoadLibrary(base_file)
         self.lib.sampling.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64]
         self.lib.getHeadBatch.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
@@ -112,7 +113,6 @@ class Config(object):
             self.trainTotal = self.lib.getTrainTotal()
             self.testTotal = self.lib.getTestTotal()
             self.validTotal = self.lib.getValidTotal()
-            print(self.relTotal, self.entTotal, self.trainTotal, self.testTotal, self.validTotal)
             self.batch_size = int(self.lib.getTrainTotal() / self.nbatches)
             self.batch_seq_size = self.batch_size * (1 + self.negative_ent + self.negative_rel)
             self.batch_h = np.zeros(self.batch_size * (1 + self.negative_ent + self.negative_rel), dtype=np.int64)
@@ -157,6 +157,7 @@ class Config(object):
 
     def set_in_path(self, path):
         self.in_path = path
+        generate_type_constraint_specification(path, exist_ok=True)
 
     def set_out_files(self, path):
         self.out_path = path
@@ -355,69 +356,40 @@ class Config(object):
                     self.ckpt.save(self.exportName + 'ckpt')
                 if self.out_path != None:
                     self.save_parameters(self.out_path)
+
     def test(self):
         with self.graph.as_default():
             with self.sess.as_default():
                 if self.importName != None:
-                        self.restore_tensorflow()
+                    self.restore_tensorflow()
                 if self.test_link_prediction:
-                        total = self.lib.getTestTotal()
-                        for times in range(total):
-                            self.lib.getHeadBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
-                            res = self.test_step(self.test_h, self.test_t, self.test_r)
-                            self.lib.testHead(res.__array_interface__['data'][0])
+                    total = self.lib.getTestTotal()
+                    for times in range(total):
+                        self.lib.getHeadBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
+                        res = self.test_step(self.test_h, self.test_t, self.test_r)
+                        self.lib.testHead(res.__array_interface__['data'][0])
 
-                            self.lib.getTailBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
-                            res = self.test_step(self.test_h, self.test_t, self.test_r)
-                            self.lib.testTail(res.__array_interface__['data'][0])
-                            if self.log_on:
-                                print(times)
-                                self.lib.test_link_prediction()
+                        self.lib.getTailBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
+                        res = self.test_step(self.test_h, self.test_t, self.test_r)
+                        self.lib.testTail(res.__array_interface__['data'][0])
+                        if self.log_on:
+                            print(times)
+
+                    # print("testing link prediction...")
+                    # print(self.lib)
+                    self.lib.test_link_prediction(True, False, True)
+                    # print("tested link prediction...")
                 if self.test_triple_classification:
-                        self.lib.getValidBatch(self.valid_pos_h_addr, self.valid_pos_t_addr, self.valid_pos_r_addr, self.valid_neg_h_addr, self.valid_neg_t_addr, self.valid_neg_r_addr)
-                        res_pos = self.test_step(self.valid_pos_h, self.valid_pos_t, self.valid_pos_r)
-                        res_neg = self.test_step(self.valid_neg_h, self.valid_neg_t, self.valid_neg_r)
-                        self.lib.getBestThreshold(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
+                    self.lib.getValidBatch(self.valid_pos_h_addr, self.valid_pos_t_addr, self.valid_pos_r_addr, self.valid_neg_h_addr, self.valid_neg_t_addr, self.valid_neg_r_addr)
+                    res_pos = self.test_step(self.valid_pos_h, self.valid_pos_t, self.valid_pos_r)
+                    res_neg = self.test_step(self.valid_neg_h, self.valid_neg_t, self.valid_neg_r)
+                    self.lib.getBestThreshold(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
 
-                        self.lib.getTestBatch(self.test_pos_h_addr, self.test_pos_t_addr, self.test_pos_r_addr, self.test_neg_h_addr, self.test_neg_t_addr, self.test_neg_r_addr)
+                    self.lib.getTestBatch(self.test_pos_h_addr, self.test_pos_t_addr, self.test_pos_r_addr, self.test_neg_h_addr, self.test_neg_t_addr, self.test_neg_r_addr)
 
-                        res_pos = self.test_step(self.test_pos_h, self.test_pos_t, self.test_pos_r)
-                        res_neg = self.test_step(self.test_neg_h, self.test_neg_t, self.test_neg_r)
-                        self.lib.test_triple_classification(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
-    # def test(self):
-    #     with self.graph.as_default():
-    #         with self.sess.as_default():
-    #             if self.importName != None:
-    #                 self.restore_tensorflow()
-    #             if self.test_link_prediction:
-    #                 total = self.lib.getTestTotal()
-    #                 for times in range(total):
-    #                     print(self.test_h_addr, self.test_t_addr, self.test_r_addr)
-    #                     print(total, self.test_h)
-    #                     self.lib.getHeadBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
-    #                     print(self.test_h, self.test_t)
-    #                     res = self.test_step(self.test_h, self.test_t, self.test_r)
-    #                     print("foo")
-    #                     self.lib.testHead(res.__array_interface__['data'][0])
-    #                     print("bar")
-
-    #                     self.lib.getTailBatch(self.test_h_addr, self.test_t_addr, self.test_r_addr)
-    #                     res = self.test_step(self.test_h, self.test_t, self.test_r)
-    #                     self.lib.testTail(res.__array_interface__['data'][0])
-    #                     if self.log_on:
-    #                         print(times)
-    #                 self.lib.test_link_prediction()
-    #             if self.test_triple_classification:
-    #                 self.lib.getValidBatch(self.valid_pos_h_addr, self.valid_pos_t_addr, self.valid_pos_r_addr, self.valid_neg_h_addr, self.valid_neg_t_addr, self.valid_neg_r_addr)
-    #                 res_pos = self.test_step(self.valid_pos_h, self.valid_pos_t, self.valid_pos_r)
-    #                 res_neg = self.test_step(self.valid_neg_h, self.valid_neg_t, self.valid_neg_r)
-    #                 self.lib.getBestThreshold(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
-
-    #                 self.lib.getTestBatch(self.test_pos_h_addr, self.test_pos_t_addr, self.test_pos_r_addr, self.test_neg_h_addr, self.test_neg_t_addr, self.test_neg_r_addr)
-
-    #                 res_pos = self.test_step(self.test_pos_h, self.test_pos_t, self.test_pos_r)
-    #                 res_neg = self.test_step(self.test_neg_h, self.test_neg_t, self.test_neg_r)
-    #                 self.lib.test_triple_classification(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
+                    res_pos = self.test_step(self.test_pos_h, self.test_pos_t, self.test_pos_r)
+                    res_neg = self.test_step(self.test_neg_h, self.test_neg_t, self.test_neg_r)
+                    self.lib.test_triple_classification(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
 
     def predict_head_entity(self, t, r, k):
         r'''This mothod predicts the top k head entities given tail entity and relation.
@@ -482,7 +454,7 @@ class Config(object):
         print(res)
         return res
 
-    def predict_triple(self, h, t, r):
+    def predict_triple(self, h, t, r, thresh=None):
         r'''This method tells you whether the given triple (h, t, r) is correct of wrong
 
         Args:
@@ -491,14 +463,21 @@ class Config(object):
             r (int): relation id
             thresh (fload): threshold for the triple
         '''
-        # self.init_triple_classification()
+        self.init_triple_classification()
         if self.importName != None:
             self.restore_tensorflow()
         res = self.test_step(np.array([h]), np.array([t]), np.array([r]))
-        return res
-
-    def predict_triples(self, triples: Tuple):
-        if self.importName != None:
-            self.restore_tensorflow()
-        res = self.test_step(np.array([triple[0] for triple in triples]), np.array([triple[1] for triple in triples]), np.array([triple[2] for triple in triples]))
-        return res
+        if thresh != None:
+            if res < thresh:
+                print("triple (%d,%d,%d) is correct" % (h, t, r))
+            else:
+                print("triple (%d,%d,%d) is wrong" % (h, t, r))
+            return
+        self.lib.getValidBatch(self.valid_pos_h_addr, self.valid_pos_t_addr, self.valid_pos_r_addr, self.valid_neg_h_addr, self.valid_neg_t_addr, self.valid_neg_r_addr)
+        res_pos = self.test_step(self.valid_pos_h, self.valid_pos_t, self.valid_pos_r)
+        res_neg = self.test_step(self.valid_neg_h, self.valid_neg_t, self.valid_neg_r)
+        self.lib.getBestThreshold(self.relThresh_addr, res_pos.__array_interface__['data'][0], res_neg.__array_interface__['data'][0])
+        if res < self.relThresh[r]:
+            print("triple (%d,%d,%d) is correct" % (h, t, r))
+        else:
+            print("triple (%d,%d,%d) is wrong" % (h, t, r))
